"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ALL_PROPERTIES_NOT_FOUND = new Error('All Properties Not Found');
const defaultObject = Object.freeze(Object.create(null));
/**
 * Uses ES2015 Proxy to partially apply a function that takes in an options object
 * @param f :: Object a => (a -> b)
 */
exports.curryObj = (f) => _curryObj(f, defaultObject);
function _curryObj(f, previousObj) {
    return (x) => {
        const obj = proxyObject(Object.assign({}, previousObj, x));
        try {
            return f(obj);
        }
        catch (error) {
            if (error !== ALL_PROPERTIES_NOT_FOUND) {
                throw error;
            }
            return _curryObj(f, obj);
        }
    };
}
function proxyObject(a) {
    return new Proxy(a, {
        get(target, key) {
            if (!target.hasOwnProperty(key)) {
                throw ALL_PROPERTIES_NOT_FOUND;
            }
            return target[key];
        },
    });
}
//# sourceMappingURL=curryObj.js.map