"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const common_1 = require("@typed/common");
/**
 * Memoize a function
 * @param f Function to memoize
 */
exports.memoize = (f) => {
    const cache = new Map();
    return (...args) => {
        const key = args.reduce((x, y) => x + common_1.toString(y), '');
        if (cache.has(key)) {
            return cache.get(key);
        }
        let result = f(...args);
        if (typeof result === 'function') {
            result = exports.memoize(result);
        }
        cache.set(key, result);
        return result;
    };
};
//# sourceMappingURL=memoize.js.map